
module grace_dict;

import grace_def;
import grace_cstk;
import grace_dict;
import grace_reader;
import std.c.windows.windows;

// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------
// -------------------- dictionary

const int FLG_F = 1;
const int FLG_M = 2;

struct header
{
  header  *nxt;  // next in list
  char[]   nam;  // name
  int      namh; // name hash
  char[][] par;  // parameter list
  int[]    parh; // parameter list hash
  char[][] ret;  // return list
  int      flg;  // flags
  cell     dat;  // data
}

header *dicthead;

int  hdr_get_flg(header *h) {return h.flg;}
void hdr_set_flg(header *h,int flg) {h.flg=flg;}
cell hdr_get_dat(header *h) {return h.dat;}
void hdr_set_dat(header *h,cell dat) {h.dat=dat;}
char[] hdr_get_nam(header *h) {return h.nam;}
void hdr_set_nam(header *h,char[] nam) {h.nam=nam;}

cell[] strarr2lst(char[][] s)
{
  cell[] l;
  for (int k=0;k<s.length;k++)
    l~=str2cell(s[k]);
  return l;
}

cell[] dict_aslist()
// reflect dictionary into list of sublists
// sublist layout: [name,flags,pars,rets,data]
{
  header *h=dicthead;
  cell[] l;
  while (h)
  {
    cell[] sl;
    sl~=str2cell(h.nam);
    sl~=int2cell(h.flg);
    sl~=lst2cell(strarr2lst(h.par));
    sl~=lst2cell(strarr2lst(h.ret));
    sl~=h.dat;
    l~=lst2cell(sl);
    h=h.nxt;
  }
  return l;
  // dpush_lst(l);
}

void dict_add(char[] nam,cell dat,int flg=0,char[][] par=[],char[][] ret=[])
{
  header *h=new header;
  h.nam=nam;
  h.namh=nam.str2uh();
  h.nxt=dicthead;
  h.flg=flg;
  h.dat=dat;
  h.par=par;
  h.parh.length=par.length;
  for (int k=0;k<par.length;k++)
    h.parh[k]=par[k].str2uh();
  h.ret=ret;
  dicthead=h;
}

void dict_addf(char[] nam,cell dat,char[][] par=[],char[][] ret=[])
{
  dict_add(nam,dat,FLG_F,par,ret);
}

void dict_addm(char[] nam,cell dat,char[][] par=[],char[][] ret=[])
{
  dict_add(nam,dat,FLG_F|FLG_M,par,ret);
}

void dict_addn(char[] nam,void* dat,int flg,char[][] par=[],char[][] ret=[])
{
  dict_add(nam,nfp2cell(dat),flg,par,ret);
}

void dict_addnf(char[] nam,void* dat,char[][] par=[],char[][] ret=[])
{
  dict_add(nam,nfp2cell(dat),FLG_F,par,ret);
}

void dict_addnm(char[] nam,void* dat,char[][] par=[],char[][] ret=[])
{
  dict_add(nam,nfp2cell(dat),FLG_F|FLG_M,par,ret);
}

void dict_show()
{
  header *h=dicthead;
  while (h)
  {
//    printf("%p ",h);
    writef("%-12s",h.nam);
//    if (h.flg&FLG_N) printf("n"); else printf(".");
    if (h.flg&FLG_F) printf("f"); else printf(".");
    if (h.flg&FLG_M) printf("i"); else printf(".");
    writef("   %s",cell_vs(h.dat));
    printf("\n");
    h=h.nxt;
  }
}

void dict_show(header *h)
{
  while (h)
  {
//    printf("%p ",h);
    writef("%-12s",h.nam);
//    if (h.flg&FLG_N) printf("n"); else printf(".");
    if (h.flg&FLG_F) printf("f"); else printf(".");
    if (h.flg&FLG_M) printf("i"); else printf(".");
    printf("   %s",cell_vs(h.dat));
    printf("\n");
    h=h.nxt;
  }
}

header* dict_find(header* h,int function(header*) check)
{
  while (h)
  {
    if (check(h))
      return h;
    h=h.nxt;
  }
  return h;
}

header* dict_find(header* h,int delegate(header*) check)
{
  while (h)
  {
    if (check(h))
      return h;
    h=h.nxt;
  }
  return h;
}

header* dict_find_untyped(header* h,char[] nam)
{
  int namh=nam.str2uh();
  while (h)
  {
    if (h.namh==namh)
    {
      if (h.nam==nam)
      {
        return h;
      }
    }
    h=h.nxt;
  }
  return h;
}

header* dict_find_typed(header* h,char[] nam)
{
//  return dict_find_typed_debug(h,nam);
  int[] parh;
  int namh=nam.str2uh();
  int ast=str2uh("*");
  while (h)
  {
    if (h.namh==namh)
    {
      if (!h.par.length)
      {
        return h;
      }
      if (stack_depth()>=h.par.length)
      {
        int ok=1;
        for (int k=0;k<h.par.length;k++)
        {
          if (k>=parh.length)
          {
            parh~=dthat(h.par.length-k-1);
          }
          if (h.parh[k]==ast)
          {
            continue;
          }
          if (parh[k]!=h.parh[k])
          {
            ok=0;
            break;
          }
        }
        if (ok)
        {
          return h;
        }
      }
    }
    h=h.nxt;
  }
  return h;
}

header* dict_find_typed_debug(header* h,char[] nam)
{
//  return dict_find_typed(h,nam);
  int[] parh;
  int namh=nam.str2uh();
  int ast=str2uh("*");
  while (h)
  {
    if (h.namh==namh)
    {
      if (!h.par.length)
      {
        return h;
      }
      // preallocate array
      parh.length=h.par.length;
      parh.length=0;
      //
      writef("\nchecking '",nam,"'; ");
      writef("pars=",h.par.length,"; ");
      writef("depth=",stack_depth(),"; ");
      if (stack_depth()>=h.par.length)
      {
        writef("length=",parh.length,"; ");
        writef("testing: ");
        int ok=1;
        for (int k=0;k<h.par.length;k++)
        {
          writef(k);
          writef("a");
          if (k>=parh.length)
          {
            writef(".");
            parh~=dthat(h.par.length-k-1);
          }
          writef("b");
          if (h.parh[k]==ast)
          {
            writef(".");
            continue;
          }
          writef("c");
          if (parh[k]!=h.parh[k])
          {
            writef(".");
            writef(" parh=",parh[k]," h.parh=",h.parh[k]," h.par=",h.par[k]);
            ok=0;
            break;
          }
        }
        writef("\n");
        if (ok)
        {
          return h;
        }
      }
    }
    h=h.nxt;
  }
  return h;
}

