
module grace_prim_hof;

import grace_def;
import grace_cstk;
import grace_dict;
import grace_eval;
import grace_reader;
import grace_prim;

void reg_prims_hof()
{
  // higher order functions
  dict_addnf("if"    ,&p_if    ,["*","*"]);
  dict_addnf("ife"   ,&p_ife   ,["*","*","*"]);
  dict_addnf("while" ,&p_while ,["*","*"]);
  dict_addnf("map"   ,&p_map   ,["lst","*"]);
  dict_addnf("each"  ,&p_each  ,["lst","*"]);
  dict_addnf("filter",&p_filter,["lst","*"]);
  dict_addnf("fold"  ,&p_fold  ,["lst","*","*"]);
  dict_addnf("dip"   ,&p_dip   ,["*","*","*"]);
  dict_addnf("infra" ,&p_infra ,["lst","*"]);
}

static this()
{
  reg_prims~=&reg_prims_hof;
}
// --------------------
// ----- higher order

void p_if()
{
  cell c=dpop(); // condifiton
  cell b=dpop(); // body
  eval(c);
  if (cell2log(dpop()))
    eval(b);
}

void p_ife()
{
  cell c=dpop(); // condifiton
  cell b=dpop(); // body
  cell a=dpop(); // altb
  eval(c);
  if (cell2log(dpop()))
    eval(b);
  else
    eval(a);
}

void p_while()
{
  cell c; // condifiton
  cell b; // body
  c=dpop();
  b=dpop();
  eval(c);
  while (cell2log(dpop()))
  {
    eval(b);
    eval(c);
  }
}

void p_map()
{
  cell   f; // function
  cell[] l; // list
  f=dpop();
  l=cell2lst(dtos());
  dtoset(lst2cell(new_lst()));
  denter();
  for (int k=0;k<l.length;k++)
  {
    dpush(l[k]);
    eval(f);
    l[k]=dpop();
  }
  dleave();
  dtoset(lst2cell(l));
}

void p_each()
{
  cell   f; // function
  cell[] l; // list
  f=dpop();
  l=dpop_lst();
  denternew();
  for (int k=0;k<l.length;k++)
  {
    dpush(l[k]);
    eval(f);
    stack_zapp();
  }
  dleave();
  dpop();
}

void p_filter()
{
  cell   f; // function
  cell[] ol; // org list
  cell[] nl; // new list
  f=dpop();
  ol=dpop_lst();
  denternew();
  for (int k=0;k<ol.length;k++)
  {
    dpush(ol[k]);
    eval(f);
    if (dpop_log())
      nl~=ol[k];
    stack_zapp();
  }
  dleave();
  dpop();
  dpush_lst(nl);
}

void p_fold()
{
  cell   f; // function
  cell   i; // init
  cell[] l; // list
  f=dpop();
  i=dpop();
  l=cell2lst(dtos());
  dtoset(lst2cell(new_lst()));
  denter();
  dpush(i);
  for (int k=0;k<l.length;k++)
  {
    dpush(l[k]);
    eval(f);
  }
  i=dpop();
  dleave();
  dtoset(i);
}

void p_dip()
{
  cell f; // function
  cell s; // save
  f=dpop();
  s=dpop();
  eval(f);
  dpush(s);
}

void p_infra()
{
  cell f; // function
  f=dpop();
  denter();
  eval(f);
  dleave();
}
